(*
  Unofficial patch for JVNVU#97910946

  Copyright (c) 2014 Lyna

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)
unit jvnvu97910946fix;

interface

implementation

uses
  Windows, SysUtils, Classes, Graphics, Consts;

var
  PaletteFromDIBColorTable: Pointer = nil;
  TMetafile_ReadEMFStream: Pointer = nil;
  TBitmap_ReadDIB: Pointer = nil;

  // For FixedPaletteFromDIBColorTable
  SystemPaletteOverride: function(var Pal: TMaxLogPalette): Boolean = nil;
  ByteSwapColors: procedure(var Colors; Count: Integer) = nil;

  // For TBitmapHelper.FixedReadDIB
  GetDInColors: function(BitCount: Word): Integer = nil;
  RGBTripleToQuad: procedure(var ColorTable) = nil;
  GDICheck: function(Value: THandle): THandle = nil;

function FixedPaletteFromDIBColorTable(DIBHandle: THandle; ColorTable: Pointer;
  ColorCount: Integer): HPalette;
var
  DC: HDC;
  Save: THandle;
  Pal: TMaxLogPalette;
begin
  Result := 0;
  Pal.palVersion := $300;
  if DIBHandle <> 0 then
  begin
    DC := CreateCompatibleDC(0);
    Save := SelectObject(DC, DIBHandle);
    Pal.palNumEntries := GetDIBColorTable(DC, 0, 256, Pal.palPalEntry);
    SelectObject(DC, Save);
    DeleteDC(DC);
  end
  else
  begin
    if ColorCount > 256 then
      raise EInvalidGraphic.CreateRes(@SInvalidBitmap);
    Pal.palNumEntries := ColorCount;
    Move(ColorTable^, Pal.palPalEntry, ColorCount * 4);
  end;
  if Pal.palNumEntries = 0 then Exit;
  if (Pal.palNumEntries <> 16) or not SystemPaletteOverride(Pal) then
    ByteSwapColors(Pal.palPalEntry, Pal.palNumEntries);
  Result := CreatePalette(PLogPalette(@Pal)^);
end;

type
{$HINTS OFF}
  TDummyMetafileImage = class(TSharedImage)
  private
    FHandle: HENHMETAFILE;
    FWidth: Integer;
    FHeight: Integer;
    FPalette: HPALETTE;
    FInch: Word;
  end;
{$HINTS ON}

  TMetafileHelper = class helper for TMetafile
  private
    procedure FixedReadEMFStream(Stream: TStream);
  end;

procedure TMetafileHelper.FixedReadEMFStream(Stream: TStream);
var
  EnhHeader: TEnhMetaheader;
  Buf: PByte;
begin
  Self.NewImage;
  Stream.ReadBuffer(EnhHeader, Sizeof(EnhHeader));
  if EnhHeader.dSignature <> ENHMETA_SIGNATURE then
    raise EInvalidGraphic.CreateRes(@SInvalidMetafile);
  if EnhHeader.nBytes < Sizeof(EnhHeader) then
    raise EInvalidGraphic.CreateRes(@SInvalidMetafile);
  GetMem(Buf, EnhHeader.nBytes);
  with TDummyMetafileImage(Self.FImage) do
  try
    Move(EnhHeader, Buf^, Sizeof(EnhHeader));
    Stream.ReadBuffer(PByte(Buf + Sizeof(EnhHeader))^,
      EnhHeader.nBytes - Sizeof(EnhHeader));
    FHandle := SetEnhMetafileBits(EnhHeader.nBytes, Buf);
    if FHandle = 0 then raise EInvalidGraphic.CreateRes(@SInvalidMetafile);
    FInch := 0;
    with EnhHeader.rclFrame do
    begin
      FWidth := Right - Left;
      FHeight := Bottom - Top;
    end;
    Enhanced := True;
  finally
    FreeMem(Buf, EnhHeader.nBytes);
  end;
end;

type
{$HINTS OFF}
  TDummyBitmapImage = class(TSharedImage)
  private
    FHandle: HBITMAP;
    FMaskHandle: HBITMAP;
    FPalette: HPALETTE;
    FDIBHandle: HBITMAP;
    FDIB: TDIBSection;
    FSaveStream: TMemoryStream;
    FOS2Format: Boolean;
    FHalftone: Boolean;
  end;
{$HINTS ON}

  TBitmapHelper = class helper for TBitmap
  private
    procedure FixedReadDIB(Stream: TStream; ImageSize: LongWord; bmf: PBitmapFileHeader);
  end;

procedure TBitmapHelper.FixedReadDIB(Stream: TStream; ImageSize: LongWord; bmf: PBitmapFileHeader);
const
  DIBPalSizes: array [Boolean] of Byte = (SizeOf(TRGBQuad), SizeOf(TRGBTriple));
var
  DC, MemDC: HDC;
  BitsMem: Pointer;
  OS2Header: TBitmapCoreHeader;
  BitmapInfo: PBitmapInfo;
  ColorTable: Pointer;
  HeaderSize: Integer;
  OS2Format: Boolean;
  BMHandle, OldBMP: HBITMAP;
  DIB: TDIBSection;
  Pal, OldPal: HPalette;
  RLEStream: TStream;
  vbmf: TBitmapFileHeader;
begin
  Pal := 0;
  BMHandle := 0;
  RLEStream := nil;
  Stream.Read(HeaderSize, SizeOf(HeaderSize));
  OS2Format := HeaderSize = SizeOf(OS2Header);
  if OS2Format then
    HeaderSize := SizeOf(TBitmapInfoHeader);
  GetMem(BitmapInfo, HeaderSize + 12 + 256 * SizeOf(TRGBQuad));
  with BitmapInfo^ do
  try
    try
      if OS2Format then
      begin
        Stream.Read(Pointer(PByte(@OS2Header) + SizeOf(HeaderSize))^,
          SizeOf(OS2Header) - SizeOf(HeaderSize));
        FillChar(bmiHeader, SizeOf(bmiHeader), 0);
        with bmiHeader, OS2Header do
        begin
          biWidth := bcWidth;
          biHeight := bcHeight;
          biPlanes := bcPlanes;
          biBitCount := bcBitCount;
        end;
        Dec(ImageSize, SizeOf(OS2Header));
      end
      else
      begin
        Stream.Read(Pointer(PByte(BitmapInfo) + SizeOf(HeaderSize))^,
          HeaderSize - SizeOf(HeaderSize));
        Dec(ImageSize, HeaderSize);

        if (bmiHeader.biCompression <> BI_BITFIELDS) and
          (bmiHeader.biCompression <> BI_RGB) then
        begin
          RLEStream := TMemoryStream.Create;
          if bmf = nil then
          begin
            FillChar(vbmf, SizeOf(vbmf), 0);
            vbmf.bfType := $4D42;
            vbmf.bfSize := ImageSize + Cardinal(HeaderSize);
            bmf := @vbmf;
          end;
          RLEStream.Write(bmf^, SizeOf(bmf^));
          RLEStream.Write(HeaderSize, SizeOf(HeaderSize));
          RLEStream.Write(Pointer(PByte(BitmapInfo) + SizeOf(HeaderSize))^,
            HeaderSize - SizeOf(HeaderSize));
          RLEStream.CopyFrom(Stream, ImageSize);
          RLEStream.Seek(-Integer(ImageSize), soFromEnd);
          Stream := RLEStream;  // the rest of the proc reads from RLEStream
        end;
      end;

      with bmiHeader do
      begin
        biSize := HeaderSize;
        ColorTable := Pointer(PByte(BitmapInfo) + HeaderSize);

        if biPlanes <> 1 then raise EInvalidGraphic.CreateRes(@SInvalidBitmap);

        if (HeaderSize = SizeOf(TBitmapInfoHeader)) and
          ((biBitCount = 16) or (biBitCount = 32)) and
          (biCompression = BI_BITFIELDS) then
        begin
          Stream.ReadBuffer(ColorTable^, 3 * SizeOf(DWORD));
          Inc(PByte(ColorTable), 3 * SizeOf(DWORD));
          Dec(ImageSize, 3 * SizeOf(DWORD));
        end;

        if biClrUsed = 0 then
          biClrUsed := GetDInColors(biBitCount);

        if (biClrUsed * DIBPalSizes[OS2Format]) > (256 * SizeOf(TRGBQuad)) then
           raise EInvalidGraphic.CreateRes(@SInvalidBitmap);

        Stream.ReadBuffer(ColorTable^, biClrUsed * DIBPalSizes[OS2Format]);
        Dec(ImageSize, biClrUsed * DIBPalSizes[OS2Format]);

        if (biSizeImage = 0) or (biCompression = BI_RGB) then // top-down DIBs have negative height
          biSizeImage := BytesPerScanLine(biWidth, biBitCount, 32) * Abs(biHeight);

        if biSizeImage < ImageSize then
          ImageSize := biSizeImage;
      end;

      if OS2Format then RGBTripleToQuad(ColorTable^);

      DC := GDICheck(GetDC(0));
      try
        if ((bmiHeader.biCompression <> BI_RGB) and
          (bmiHeader.biCompression <> BI_BITFIELDS)) or DDBsOnly then
        begin
          MemDC := 0;
          GetMem(BitsMem, ImageSize);
          try
            Stream.ReadBuffer(BitsMem^, ImageSize);
            MemDC := GDICheck(CreateCompatibleDC(DC));
            OldBMP := SelectObject(MemDC, CreateCompatibleBitmap(DC, 1, 1));
            OldPal := 0;
            if bmiHeader.biClrUsed > 0 then
            begin
              Pal := FixedPaletteFromDIBColorTable(0, ColorTable, bmiHeader.biClrUsed);
              OldPal := SelectPalette(MemDC, Pal, False);
              RealizePalette(MemDC);
            end;

            try
              BMHandle := CreateDIBitmap(MemDC, BitmapInfo^.bmiHeader, CBM_INIT, BitsMem,
                BitmapInfo^, DIB_RGB_COLORS);
              if (BMHandle = 0) then
                if GetLastError = 0 then
                  raise EInvalidGraphic.CreateRes(@SInvalidBitmap)
                else RaiseLastOSError;
            finally
              if OldPal <> 0 then
                SelectPalette(MemDC, OldPal, True);
              DeleteObject(SelectObject(MemDC, OldBMP));
            end;
          finally
            if MemDC <> 0 then DeleteDC(MemDC);
            FreeMem(BitsMem);
          end;
        end
        else
        begin
          BMHandle := CreateDIBSection(DC, BitmapInfo^, DIB_RGB_COLORS, BitsMem, 0, 0);
          if (BMHandle = 0) or (BitsMem = nil) then
            if GetLastError = 0 then
              raise EInvalidGraphic.CreateRes(@SInvalidBitmap)
            else RaiseLastOSError;

          try
            Stream.ReadBuffer(BitsMem^, ImageSize);
          except
            DeleteObject(BMHandle);
            raise;
          end;
        end;
      finally
        ReleaseDC(0, DC);
      end;

      if (bmiHeader.biBitCount > 8) and (bmiHeader.biClrUsed > 0) and (Pal = 0)then
        Pal := FixedPaletteFromDIBColorTable(0, ColorTable, bmiHeader.biClrUsed);

      FillChar(DIB, SizeOf(DIB), 0);
      GetObject(BMHandle, Sizeof(DIB), @DIB);
      DIB.dsBmih.biXPelsPerMeter := bmiHeader.biXPelsPerMeter;
      DIB.dsBmih.biYPelsPerMeter := bmiHeader.biYPelsPerMeter;
      DIB.dsBmih.biClrUsed := bmiHeader.biClrUsed;
      DIB.dsBmih.biClrImportant := bmiHeader.biClrImportant;
    except
      RLEStream.Free;
      raise;
    end;
  finally
    FreeMem(BitmapInfo);
  end;
  Self.NewImage(BMHandle, Pal, DIB, OS2Format, RLEStream);

  if (TDummyBitmapImage(Self.FImage).FDIB.dsBMIh.biBitCount = 32) and (Self.FAlphaFormat = afDefined) then
    Self.PreMultiplyAlpha;

  PaletteModified := Palette <> 0;
  Changed(Self);
end;

procedure Patch(OldAddr, NewAddr: Pointer);
type
  TJmp = packed record
    OpCode: Byte;
    Offset: Longint;
  end;
  PJmp = ^TJmp;
const
  OP_JMP = $E9;
var
  oldProtect: DWORD;
begin
  VirtualProtect(OldAddr, SizeOf(TJmp), PAGE_EXECUTE_READWRITE, oldProtect);
  PJmp(OldAddr)^.OpCode := OP_JMP;
  PJmp(OldAddr)^.Offset := NativeInt(NewAddr) - NativeInt(OldAddr) - SizeOf(TJmp);
  VirtualProtect(OldAddr, SizeOf(TJmp), oldProtect, oldProtect);
  FlushInstructionCache(GetCurrentProcess, OldAddr, SizeOf(TJmp));
end;

initialization
  if PaletteFromDIBColorTable = nil then Exit;
  if TMetafile_ReadEMFStream = nil then Exit;
  if TBitmap_ReadDIB = nil then Exit;

  Patch(PaletteFromDIBColorTable, @FixedPaletteFromDIBColorTable);
  Patch(TMetafile_ReadEMFStream, @TMetafile.FixedReadEMFStream);
  Patch(TBitmap_ReadDIB, @TBitmap.FixedReadDIB);
end.
